@@ -91,7 +91,8 @@ get_icon_for_device(const char *mountpoint)
 	if (!volume) {
 		g_warning("Cannot get volume for mount point '%s'",
 			  mountpoint);
-		return NULL;
+		/* Fallback using / icon */
+		volume = gnome_vfs_volume_monitor_get_volume_for_path(monitor, "/");
 	}
 
 	icon_name = gnome_vfs_volume_get_icon(volume);
@@ -105,6 +106,71 @@ get_icon_for_device(const char *mountpoint)
 }
 
 
+static gboolean
+find_disk_in_model(GtkTreeModel *model, const char *mountpoint,
+		   GtkTreeIter *result)
+{
+	GtkTreeIter iter;
+	gboolean found = FALSE;
+
+	if (gtk_tree_model_get_iter_first(model, &iter)) {
+		do {
+			char *dir;
+
+			gtk_tree_model_get(model, &iter,
+					   DISK_DIR, &dir,
+					   -1);
+
+			if (dir && !strcmp(dir, mountpoint)) {
+				*result = iter;
+				found = TRUE;
+			}
+
+			g_free(dir);
+
+		} while (!found && gtk_tree_model_iter_next(model, &iter));
+	}
+
+	return found;
+}
+
+
+
+static void
+remove_old_disks(GtkTreeModel *model, const glibtop_mountentry *entries, guint n)
+{
+	GtkTreeIter iter;
+
+	if (!gtk_tree_model_get_iter_first(model, &iter))
+		return;
+
+	do {
+		char *dir;
+		guint i;
+		gboolean found = FALSE;
+
+		gtk_tree_model_get(model, &iter,
+				   DISK_DIR, &dir,
+				   -1);
+
+		for (i = 0; i != n; ++i) {
+			if (!strcmp(dir, entries[i].mountdir)) {
+				found = TRUE;
+				break;
+			}
+		}
+
+		if (!found) {
+			if (!gtk_list_store_remove(GTK_LIST_STORE(model), &iter))
+				break;
+		}
+
+		g_free(dir);
+
+	} while (gtk_tree_model_iter_next(model, &iter));
+}
+
+
 
 static void
 add_disk(GtkListStore *list, const glibtop_mountentry *entry)
@@ -126,7 +192,13 @@ add_disk(GtkListStore *list, const glibtop_mountentry *entry)
 	avail_str = SI_gnome_vfs_format_file_size_for_display(bavail);
 	total_str = SI_gnome_vfs_format_file_size_for_display(btotal);
 
-	gtk_list_store_append(list, &iter);
+	/* if we can find a row with the same mountpoint, we get it but we
+	   still need to update all the fields.
+	   This makes selection persistent.
+	*/
+	if (!find_disk_in_model(GTK_TREE_MODEL(list), entry->mountdir, &iter))
+		gtk_list_store_append(list, &iter);
+
 	gtk_list_store_set(list, &iter,
 			   DISK_ICON, pixbuf,
 			   DISK_DEVICE, entry->devname,
@@ -165,10 +237,11 @@ cb_update_disks(gpointer data)
 	guint i;
 
 	list = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(procdata->disk_list)));
-	gtk_list_store_clear(list);
 
 	entries = glibtop_get_mountlist(&mountlist, procdata->config.show_all_fs);
 
+	remove_old_disks(GTK_TREE_MODEL(list), entries, mountlist.number);
+
 	for (i = 0; i < mountlist.number; i++)
 		add_disk(list, &entries[i]);
 
@@ -188,6 +261,37 @@ cb_disk_columns_changed(GtkTreeView *treeview, gpointer user_data)
 				"/apps/procman/disktreenew");
 }
 
+
+static void open_dir(GtkTreeView       *tree_view,
+		     GtkTreePath       *path,
+		     GtkTreeViewColumn *column,
+		     gpointer	       user_data)
+{
+	GtkTreeIter iter;
+	GtkTreeModel *model;
+	char *dir, *url;
+
+	model = gtk_tree_view_get_model(tree_view);
+
+	if (!gtk_tree_model_get_iter(model, &iter, path)) {
+		char *p;
+		p = gtk_tree_path_to_string(path);
+		g_warning("Cannot get iter for path '%s'\n", p);
+		g_free(p);
+		return;
+	}
+
+	gtk_tree_model_get(model, &iter, DISK_DIR, &dir, -1);
+
+	url = g_strdup_printf("file://%s", dir);
+
+	if (gnome_vfs_url_show(url) != GNOME_VFS_OK)
+		g_warning("Cannot open '%s'\n", url);
+
+	g_free(url);
+	g_free(dir);
+}
+
 GtkWidget *
 create_disk_view(ProcData *procdata)
 {
@@ -198,7 +302,7 @@ create_disk_view(ProcData *procdata)
 	GtkListStore *model;
 	GtkTreeViewColumn *col;
 	GtkCellRenderer *cell;
-	gint i;
+	guint i;
 
 	const gchar * const titles[] = {
 		N_("Device"),
@@ -249,6 +353,7 @@ create_disk_view(ProcData *procdata)
 		);
 
 	disk_tree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(model));
+	g_signal_connect(G_OBJECT(disk_tree), "row-activated", G_CALLBACK(open_dir), NULL);
 	procdata->disk_list = disk_tree;
 	gtk_container_add(GTK_CONTAINER(scrolled), disk_tree);
 	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(disk_tree), TRUE);
